import json

from functools import partial
from typing import Any, Dict, Optional

from ujenkins.exceptions import JenkinsNotFoundError


class Jobs:

    def __init__(self, jenkins) -> None:
        self.jenkins = jenkins

    def get(self, url: str = '', depth: Optional[int] = None) -> Dict[str, dict]:
        """
        Get jobs in selected folder, by default root is used.

        Args:
            url (str):
                URL from job property, by default root is used.

            depth (Optional[int]):
                Get jobs recursively including folders from selected URL using
                depth value, default is None which means no recursion.

        Example:

        .. code-block:: python

            {'folder': {'_class': 'com.cloudbees.hudson.plugins.folder.Folder',
                        'name': 'folder',
                        'url': 'http://localhost/job/folder/'},
             'project': {'_class': 'hudson.model.FreeStyleProject',
                         'color': 'blue',
                         'name': 'project',
                         'url': 'http://localhost/job/project/'}}

        Returns:
            Dict[str, dict]: full name and job properties.
        """
        def callback(response):
            all_jobs = {}
            jobs = json.loads(response.body)['jobs']
            for job in jobs:
                all_jobs[job['name']] = job

            return all_jobs

        params = None

        if depth:
            tree = ',jobs[name,url,color'*depth + ']'*depth
            params = {'tree': tree.lstrip(',')}

        return self.jenkins._request(
            'GET',
            url + '/api/json',
            params=params,
            _callback=callback,
        )

    def get_info(self, name: str) -> dict:
        """
        Get detailed information of specified job.

        Args:
            name (str):
                Job name.

        Returns:
            dict: job information.
        """
        folder_name, job_name = self.jenkins._get_folder_and_job_name(name)

        return self.jenkins._request(
            'GET',
            f'/{folder_name}/job/{job_name}/api/json'
        )

    def get_config(self, name: str) -> str:
        """
        Get XML config of a specified job.

        Args:
            name (str):
                Job name.

        Returns:
            str: XML config
        """
        folder_name, job_name = self.jenkins._get_folder_and_job_name(name)

        return self.jenkins._request(
            'GET',
            f'/{folder_name}/job/{job_name}/config.xml',
            _callback=self.jenkins._return_body,
        )

    def is_exists(self, name: str) -> bool:
        """
        Checks if the job exists.

        Args:
            name (str):
                Job name.

        Returns:
            bool: job exists.
        """
        if name == '':
            return False

        def callback1(_) -> Any:
            return partial(self.get_info, name)

        def callback2(response: Any) -> bool:
            if isinstance(response, JenkinsNotFoundError):
                return False

            return True

        return self.jenkins._chain([callback1, callback2])

    def create(self, name: str, config: str) -> None:
        """
        Create new jenkins job.

        Args:
            name (str):
                Job name.

            config (str):
                XML config of new job. It`s convenient way to use `get_config()`
                to get existing job config and change it on your taste, or to
                use `construct_config()` method.

        Returns:
            None
        """
        folder_name, job_name = self.jenkins._get_folder_and_job_name(name)

        headers = {'Content-Type': 'text/xml'}
        params = {'name': job_name}

        return self.jenkins._request(
            'POST',
            f'/{folder_name}/createItem',
            params=params,
            data=config,
            headers=headers
        )

    def reconfigure(self, name: str, config: str) -> None:
        """
        Reconfigure specified job name.

        Args:
            name (str):
                Job name or path (within folder).

            config (str):
                XML config of new job. It`s convenient way to use `get_config()`
                to get existing job config and change it on your taste, or to
                use `construct_config()` method.

        Returns:
            None
        """
        folder_name, job_name = self.jenkins._get_folder_and_job_name(name)

        return self.jenkins._request(
            'POST',
            f'/{folder_name}/job/{job_name}/config.xml',
            data=config,
            headers={'Content-Type': 'text/xml'},
        )

    def delete(self, name: str) -> None:
        """
        Delete existed jenkins job.

        Args:
            name (str):
                Job name. For job in folder just use `/`.

        Returns:
            None
        """
        folder_name, job_name = self.jenkins._get_folder_and_job_name(name)

        return self.jenkins._request(
            'POST',
            f'/{folder_name}/job/{job_name}/doDelete'
        )

    def copy(self, name: str, new_name: str) -> None:
        """
        Copy specified job.

        Args:
            name (str):
                Job name or path (within folder).

            new_name (str):
                New job name.

        Returns:
            None
        """
        folder_name, job_name = self.jenkins._get_folder_and_job_name(name)

        params = {
            'mode': 'copy',
            'from': job_name,
            'name': new_name,
        }

        return self.jenkins._request(
            'POST',
            f'/{folder_name}/createItem',
            params=params
        )

    def rename(self, name: str, new_name: str) -> None:
        """
        Rename specified job name.

        Args:
            name (str):
                Job name or path (within folder).

            new_name (str):
                New job name.

        Returns:
            None
        """
        folder_name, job_name = self.jenkins._get_folder_and_job_name(name)

        params = {
            'newName': new_name
        }

        return self.jenkins._request(
            'POST',
            f'/{folder_name}/job/{job_name}/doRename',
            params=params
        )

    def enable(self, name: str) -> None:
        """
        Enable specified job.

        Args:
            name (str):
                Job name.

        Returns:
            None
        """
        folder_name, job_name = self.jenkins._get_folder_and_job_name(name)

        return self.jenkins._request(
            'POST',
            f'/{folder_name}/job/{job_name}/enable'
        )

    def disable(self, name: str) -> None:
        """
        Disable specified job.

        Args:
            name (str):
                Job name.

        Returns:
            None
        """
        folder_name, job_name = self.jenkins._get_folder_and_job_name(name)

        return self.jenkins._request(
            'POST',
            f'/{folder_name}/job/{job_name}/disable'
        )



import json

from typing import Dict

from ujenkins.exceptions import JenkinsError


class Views:

    def __init__(self, jenkins) -> None:
        self.jenkins = jenkins

    def get(self) -> Dict[str, dict]:
        """
        Get views and their details.

        Returns:
            Dict[str, dict]: plugin name and plugin properties.
        """
        def callback(response):
            views = json.loads(response.body)['views']
            return {v['name']: v for v in views}

        return self.jenkins._request(
            'GET',
            '/api/json',
            _callback=callback,
        )

    def is_exists(self, name: str) -> bool:
        """
        Check if view exists.

        Args:
            name (str):
                View name.

        Returns:
            bool: view existing.
        """
        def callback1(_):
            return self.get

        def callback2(response):
            return name in response

        return self.jenkins._chain([callback1, callback2])

    def get_config(self, name: str) -> str:
        """
        Return view config in XML format.

        Args:
            name (str):
                View name.

        Returns:
            str: XML config of view.
        """
        return self.jenkins._request(
            'GET',
            f'/view/{name}/config.xml',
            _callback=self.jenkins._return_body,
        )

    def create(self, name: str, config: str) -> None:
        """
        Create view using XML config.

        Args:
            name (str):
                View name.

            config (str):
                XML config.

        Returns:
            None
        """
        def callback1(_):
            return self.get

        def callback2(response: dict) -> None:
            if name in response:
                raise JenkinsError(f'View `{name}` is already exists')

            headers = {'Content-Type': 'text/xml'}
            params = {'name': name}

            return self.jenkins._request(
                'POST',
                '/createView',
                data=config,
                params=params,
                headers=headers,
            )

        return self.jenkins._chain([callback1, callback2])

    def reconfigure(self, name: str, config: str) -> None:
        """
        Reconfigure view using XML config.

        Args:
            name (str):
                View name.

            config (str):
                XML config.

        Returns:
            None
        """
        return self.jenkins._request(
            'POST',
            f'/view/{name}/config.xml',
            data=config,
            headers={'Content-Type': 'text/xml'},
        )

    def delete(self, name: str) -> None:
        """
        Delete view.

        Args:
            name (str):
                View name.

        Returns:
            None
        """
        return self.jenkins._request('POST', f'/view/{name}/doDelete')



class JenkinsError(Exception):
    """
    Core library exception
    """
    def __init__(self, message=None, status=None):
        super().__init__(message)
        self.message = message
        self.status = status


class JenkinsNotFoundError(JenkinsError):
    """
    Raises when request return HTTP code 404 (not found)
    """



import json

from typing import Dict


class Queue:

    def __init__(self, jenkins) -> None:
        self.jenkins = jenkins

    def get(self) -> Dict[int, dict]:
        """
        Get server queue.

        Returns:
            Dict[int, dict]: id item in queue, and it's detailed information.
        """
        def callback(response):
            items = json.loads(response.body)['items']
            return {item['id']: item for item in items}

        return self.jenkins._request(
            'GET',
            '/queue/api/json',
            _callback=callback,
        )

    def get_info(self, item_id: int) -> dict:
        """
        Get info about enqueued item (build) identifier.

        Args:
            item_id (int):
                enqueued item identifier.

        Returns:
            dict: identifier information.
        """
        return self.jenkins._request(
            'GET',
            f'/queue/item/{item_id}/api/json',
        )

    def cancel(self, item_id: int) -> None:
        """
        Cancel enqueued item (build) identifier.

        Args:
            item_id (int):
                enqueued item identifier.

        Returns:
            None
        """
        return self.jenkins._request(
            'POST',
            '/queue/cancelItem',
            params={'id': item_id},
        )



import asyncio

from http import HTTPStatus
from typing import Any, Callable, List, Optional, Union

from aiohttp import (
    BasicAuth,
    ClientError,
    ClientResponse,
    ClientSession,
    ClientTimeout,
    CookieJar,
)

from ujenkins.adapters import CRUMB_ISSUER_URL
from ujenkins.core import Jenkins, JenkinsError, JenkinsNotFoundError, Response


class RetryClientSession:

    def __init__(self, options: dict) -> None:
        self.total = options['total']
        self.factor = options.get('factor', 1)
        self.statuses = options.get('statuses', [])
        self.methods = options.get('methods', [
            'DELETE', 'GET', 'HEAD', 'OPTIONS', 'PUT', 'TRACE'
        ])

        self.session = ClientSession(cookie_jar=CookieJar(unsafe=True))

    async def request(self, *args: Any, **kwargs: Any) -> ClientResponse:
        for total in range(self.total):
            try:
                response = await self.session.request(*args, **kwargs)
            except (ClientError, asyncio.TimeoutError) as e:
                if total + 1 == self.total:
                    raise JenkinsError from e
            else:
                if response.method.upper() not in self.methods:
                    break
                if response.status not in self.statuses:
                    break

            await asyncio.sleep(self.factor * (2 ** (total - 1)))

        return response

    async def close(self) -> None:
        await self.session.close()


class AsyncJenkinsClient(Jenkins):

    session = None  # type: Union[ClientSession, RetryClientSession]
    timeout = None

    def __init__(self,
                 url: str,
                 user: Optional[str] = None,
                 password: Optional[str] = None,
                 *,
                 verify: bool = True,
                 timeout: Optional[float] = None,
                 retry: Optional[dict] = None
                 ) -> None:
        """
        Jenkins async client class.

        Args:
            url (str):
                URL of Jenkins server.

            user (Optional[str]):
                User name.

            password (Optional[str]):
                Password for user.

            verify (Optional[bool]):
                Verify SSL (default: true).

            timeout (Optional[int]):
                HTTP request timeout.

            retry (Optional[dict]):
                Retry options to prevent failures if server restarting or
                temporary network problem. Disabled by default use total > 0
                to enable.

                - total: ``int`` Total retries count.
                - factor: ``int`` Sleep between retries (default 1)
                    {factor} * (2 ** ({number of total retries} - 1))
                - statuses: ``List[int]`` HTTP statues retries on. (default [])
                - methods: ``List[str]`` list of HTTP methods to retry, idempotent
                    methods are used by default.

                Example:

                .. code-block:: python

                    retry = dict(
                        total=10,
                        factor=1,
                        statuses=[500]
                    )

        Returns:
            AsyncClient instance
        """
        super().__init__()

        self.host = url.rstrip('/')
        self.crumb = None  # type: Any

        self.auth = None
        if user and password:
            self.auth = BasicAuth(user, password)

        # use unsafe cookie jar to be the same as requests package and add
        # possibility to use password instead of tokens
        # https://github.com/pbelskiy/ujenkins/issues/11

        if retry:
            self._validate_retry_argument(retry)
            self.session = RetryClientSession(retry)
        else:
            self.session = ClientSession(cookie_jar=CookieJar(unsafe=True))

        self.verify = verify

        if timeout:
            self.timeout = ClientTimeout(total=timeout)

    async def _http_request(self,
                            method: str,
                            path: str,
                            *,
                            _callback: Optional[Callable] = None,
                            **kwargs: Any
                            ) -> Any:

        if self.timeout and 'timeout' not in kwargs:
            kwargs['timeout'] = self.timeout

        if self.crumb:
            kwargs.setdefault('headers', {})
            kwargs['headers'].update(self.crumb)

        if path.startswith('http'):
            url = path
        else:
            url = self.host + path

        response = await self.session.request(
            method,
            url,
            auth=self.auth,
            ssl=self.verify,
            **kwargs
        )

        body = await response.text()

        result = self._process(
            Response(response.status, response.headers, body),
            _callback
        )

        return result

    async def _get_crumb(self) -> Union[bool, dict]:
        try:
            response = await self._http_request('GET', CRUMB_ISSUER_URL)
            self.crumb = {response['crumbRequestField']: response['crumb']}
            return self.crumb
        except JenkinsNotFoundError:
            return False

    async def _request(self,
                       method: str,
                       path: str,
                       **kwargs: Any
                       ) -> Any:
        """
        Core class method for endpoints, which wraps auto crumb detection.
        """
        if self.crumb:
            try:
                return await self._http_request(method, path, **kwargs)
            except JenkinsError as e:
                if e.status != HTTPStatus.FORBIDDEN:
                    raise

        if self.crumb is not False:
            self.crumb = await self._get_crumb()

        return await self._http_request(method, path, **kwargs)

    @staticmethod
    async def _chain(functions: List[Callable]) -> Any:
        """
        Helper function for creating call chain of async and sync functions.
        """
        prev = None

        for func in functions:
            try:
                prev = func(prev)

                while True:
                    if asyncio.iscoroutine(prev):
                        prev = await prev  # type: ignore
                    elif callable(prev):
                        prev = prev()
                    else:
                        break
            except JenkinsError as e:
                prev = e

        return prev

    async def close(self) -> None:  # type: ignore
        """
        Close client session
        """
        await self.session.close()



import json

from collections import namedtuple
from http import HTTPStatus
from typing import Any, Callable, Optional, Tuple

from ujenkins.endpoints import (
    Builds,
    Jobs,
    Nodes,
    Plugins,
    Queue,
    System,
    Views,
)
from ujenkins.exceptions import JenkinsError, JenkinsNotFoundError

Response = namedtuple('Response', ['status', 'headers', 'body'])


class Jenkins:

    def __init__(self) -> None:
        self.builds = Builds(self)
        self.jobs = Jobs(self)
        self.nodes = Nodes(self)
        self.plugins = Plugins(self)
        self.queue = Queue(self)
        self.system = System(self)
        self.views = Views(self)

    @staticmethod
    def _process(response: Response, callback: Optional[Callable] = None) -> Any:
        if response.status == HTTPStatus.NOT_FOUND:
            raise JenkinsNotFoundError(response.body)

        if response.status >= HTTPStatus.BAD_REQUEST:
            if response.status in (
                    HTTPStatus.UNAUTHORIZED,
                    HTTPStatus.FORBIDDEN,
                    HTTPStatus.INTERNAL_SERVER_ERROR
            ):
                details = 'probably authentication problem:\n\n' + response.body
            else:
                details = '\n\n' + response.body

            raise JenkinsError(
                f'Request error [{response.status}], {details}',
                status=response.status,
            )

        # TODO: add response type annotations, parse json for callback
        if callback:
            return callback(response)

        if 'application/json' in response.headers.get('Content-Type', ''):
            return json.loads(response.body)

        return None

    @staticmethod
    def _get_folder_and_job_name(name: str) -> Tuple[str, str]:
        parts = name.split('/')

        job_name = parts[-1]
        folder_name = ''

        for folder in parts[:-1]:
            folder_name += f'job/{folder}/'

        return folder_name, job_name

    @staticmethod
    def _validate_retry_argument(retry: dict) -> None:
        for key in retry:
            if key not in ('total', 'factor', 'statuses'):
                raise JenkinsError('Unknown key in retry argument: ' + key)

        if retry.get('total', 0) <= 0:
            raise JenkinsError('Invalid `total` in retry argument must be > 0')

    @staticmethod
    def _return_body(response: Response) -> str:
        return response.body
